from lovely.pyrest.rest import RestService, rpcmethod_route
from lovely.pyrest.validation import validate, ValidationException


ARTICLES = {}


# Create a schema to validate the body
schema = {
    "type": "object",
    "properties": {
        "title": {
            "type": "string"
        },
        "content": {
            "type": "string",
            "required": False
        },
    }
}


@RestService('article')
class ArticlesService(object):
    """ Create and read articles """

    def __init__(self, request):
        self.request = request

    @rpcmethod_route(request_method='POST')
    @validate(schema)
    def post(self, data):
        id = len(ARTICLES)
        ARTICLES[id] = data
        return {'stored': id}

    @rpcmethod_route(route_suffix='/{id}')
    def get(self, id):
        return ARTICLES.get(id)


def bad_request(exc, request):
    request.response.status = 400
    return {
            'status': "ERROR",
            'reason': exc.message
            }


def includeme(config):
    config.add_route('article', '/article', static=True)
    config.add_view(bad_request, renderer='json',
                    context=ValidationException)



from lovely.pyrest.rest import RestService, rpcmethod_route
from lovely.pyrest.validation import validate

# Create a schema to validate the body
schema = {
    "type": "object",
    "properties": {
        "title": {
            "type": "string"
        }
    }
}


@RestService('article')
class ArticleService(object):
    """ Create and read articles """

    @rpcmethod_route(request_method='POST',
                     content_type='application/json')
    def create(self, data):
        """ Update article """
        return {'created': '123'}

    @rpcmethod_route(request_method='GET',
                     accept='application/json')
    @validate(schema)
    def get(self, id):
        """ Returns the article """
        return {'title': 'the title'}


def includeme(config):
    config.add_route('article', '/article', static=True)



import inspect
import logging
import venusian
from pyramid.config import predicates

log = logging.getLogger(__name__)

SERVICES = {}


def get_services():
    return SERVICES.values()


class ViewMapper(object):
    """ Mapper to pass request specific data to view method
    """
    def __init__(self, **kw):
        self.attr = kw.get('attr')

    def __call__(self, view):
        def wrapper(context, request):
            def json_body(request):
                """ Failsave function to access request.json_body
                """
                try:
                    return request.json_body
                except:
                    return {}

            def mapply(func, request):
                """ This function passes request.matchdict, request.params and
                request.json_body as kwargs to the given function
                """
                kw = {}
                # add kwargs from matchdict
                kw.update(request.matchdict)

                # add kwargs from request params
                kw.update(dict(request.params.items()))

                # add kwargs from request body
                kw.update(json_body(request))

                return func(**kw)

            inst = view(request)
            meth = getattr(inst, self.attr)
            return mapply(meth, request)
        return wrapper


class BaseRouteNotFoundException(Exception):
    """ A exception to indicate that the required base route was not found

    Possible reasons for this exception:
        - the route has not been defined
        - the config has not been commited before calling config.scan()
    """


class RestService(object):
    """ Decorator for REST API classes

    @RestService('users')
    class UserService(object):

        def __init__(self, request):
            self.request = request

        @rpcmethod(route_suffix='/{id}', request_method='PUT')
        def edit(self, id, data):
            # code goes here

    def includeme(config):
        config.add_route('users', '/users', static=True)
    """

    venusian = venusian

    def reverse_engineer_route(self, route):
        kw = {}
        if route.factory:
            kw['factory'] = route.factory
        if route.pregenerator:
            kw[pregenerator] = route.pregenerator

        def xhr(p):
            kw['xhr'] = p.val

        def path_info(p):
            kw['path_info'] = p.val.pattern

        def request_param(p):
            kw['request_param'] = p.val[0]

        def header(p):
            kw['header'] = p.text().split(" ", 1)[-1]

        def accept(p):
            kw['accept'] = p.val

        def custom_predicates(p):
            if not 'custom_predicates' in kw:
                kw['custom_predicates'] = []
            kw['custom_predicates'].append(p.func)

        def request_method(p):
            kw['request_method'] = p.val[0]
        predicate_map = {predicates.XHRPredicate: xhr,
                         predicates.PathInfoPredicate: path_info,
                         predicates.RequestParamPredicate: request_param,
                         predicates.HeaderPredicate: header,
                         predicates.AcceptPredicate: accept,
                         predicates.CustomPredicate: custom_predicates,
                         predicates.RequestMethodPredicate: request_method,
                        }
        for p in route.predicates:
            predicate_map[p.__class__](p)
        return kw

    def __init__(self, baseRouteName, **view_kwargs):
        self.baseRouteName = baseRouteName
        self.view_kwargs = view_kwargs
        # All methods of the services get registered here for sphinx autodoc
        self.methods = []

    def __call__(self, wrapped):
        def callback(context, name, service):
            config = context.config.with_package(info.module)
            # load the base route to get it's resolved pattern
            mapper = config.get_routes_mapper()
            baseRoute = mapper.get_route(self.baseRouteName)
            if baseRoute is None:
                raise BaseRouteNotFoundException
            # get default route arguments
            route_defaults = self.reverse_engineer_route(baseRoute)

            # get all rpcmethod decorated members
            def isRESTMethod(obj):
                return (inspect.ismethod(obj)
                        and (hasattr(obj, '__rpc_method_route__') or
                             hasattr(obj, '__rpc_method_view__'))
                    )
            methods = inspect.getmembers(service, isRESTMethod)
            # register the service
            SERVICES[self.baseRouteName] = self
            self.description = service.__doc__
            # if the module is used multiple times for documentation generation
            # the service get registered a few times so reset methods here.
            self.methods = []
            # loop through all decorated methods and add a route and a view
            # for it
            for (name, method) in methods:
                route_kw = {}
                route_kw.update(route_defaults)
                if hasattr(method, '__rpc_method_route__'):
                    route_kw.update(method.__rpc_method_route__)
                # allow http method GET by default
                if 'request_method' not in route_kw:
                    route_kw['request_method'] = 'GET'
                view_kw = {}
                view_kw.update(self.view_kwargs)
                if hasattr(method, '__rpc_method_view__'):
                    view_kw.update(method.__rpc_method_view__)
                route_name = '.'.join((service.__name__.lower(), name))
                pattern = baseRoute.pattern + route_kw.pop('route_suffix', '')
                # Register method
                validator = None
                if method.im_func.__name__ == 'validation_wrapper':
                    # index 2 of func_closure is the schema param of the validate
                    # method in the tuple, not accessible via keyword
                    validator = method.im_func.func_closure[2].cell_contents
                self.methods.append((pattern, route_kw, view_kw, method, validator))
                config.add_route(route_name, pattern, **route_kw)
                config.add_view(view=service,
                                route_name=route_name,
                                attr=name,
                                mapper=ViewMapper,
                                renderer='json',
                                **view_kw)
                log.debug('Adding REST method %s %s (%s.%s)',
                          route_kw['request_method'], pattern, service.__name__,
                          name)

        info = self.venusian.attach(wrapped, callback, category='restservice',
                                    depth=1)
        return wrapped


def rpcmethod_route(context_factory=None, **kwargs):
    """ Decorator to mark methods of classes decorated with `RestService`
    as member of the REST Service
    """

    def wrapper(f):
        f.context_factory = context_factory
        f.__rpc_method_route__ = kwargs
        return f
    return wrapper


def rpcmethod_view(context_factory=None, **kwargs):
    """ Decorator to mark methods of classes decorated with `RestService`
    as member of the REST Service
    """

    def wrapper(f):
        f.context_factory = context_factory
        f.__rpc_method_view__ = kwargs
        return f
    return wrapper



from __future__ import absolute_import
from docutils.core import publish_from_doctree, publish_doctree
from contextlib import contextmanager
from lovely.pyrest.rest import SERVICES
from pyramid import testing
from webtest import TestApp
from importlib import import_module
import unittest
import doctest
import os
import pprint
import json
import sys
import tempfile
import shutil
import sphinx.cmdline


project_root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
here = os.path.dirname(__file__)


DOC_CONF = """
extensions = ['lovely.pyrest.sphinx']
source_suffix = '.txt'
master_doc = 'index'
html_theme = 'pyramid'
"""


@contextmanager
def _temp_dir():
    tmp = tempfile.mkdtemp()
    yield tmp
    shutil.rmtree(tmp)


def render_doc(rst_path):
    """ Renders a restructured Text file and returns the
        html output """

    # Remove all existing services
    SERVICES.clear()
    with _temp_dir() as tmp:
        out_dir = os.path.join(tmp, 'out')
        in_dir = os.path.join(tmp, 'in')
        os.mkdir(in_dir)
        # write the rst string to a file
        conf = DOC_CONF % dict(
            rst_file=rst_path
        )
        shutil.copyfile(rst_path, os.path.join(in_dir, 'index.txt'))
        conf_path = os.path.join(in_dir, 'conf.py')
        with file(conf_path, 'wb') as cf:
            cf.write(conf)
        sphinx_argv = ['sphinx-build', '-b', 'html',
                       in_dir, out_dir]
        sphinx.cmdline.main(sphinx_argv)
        index_path = os.path.join(out_dir, 'index.html')
        with open(index_path, 'r') as index:
            return index.read()


def print_json(js):
    try:
        d = json.loads(js)
    except ValueError:
        print >> sys.stderr, js
        raise
    print(json.dumps(d, indent=4, sort_keys=True))


def render_doc_node(node, writer_name='pseudoxml'):
    """ Renderers a docutils node """
    # Create an empty document
    doc = publish_doctree("")
    # append the node
    doc.children.append(node)
    # publish the document and return the output
    return publish_from_doctree(doc, writer_name=writer_name)


def get_app(module_str):
    # scans the defined module and returns
    # a TestApp

    # Settings can also be defined in a .ini file.
    settings = {
        'lovely.pyrest.jsonp.param_name': 'callback'
    }
    config = testing.setUp(settings=settings)
    module = import_module(module_str)
    # check if the module has an `includeme` method and call it
    # because the base route must be added
    if hasattr(module, 'includeme'):
        module.includeme(config)
    config.commit()
    config.scan(module)
    return TestApp(config.make_wsgi_app())


def setUp(test):
    test.globs['print_json'] = print_json
    test.globs['pprint'] = pprint.pprint
    test.globs['render_doc_node'] = render_doc_node
    test.globs['render_doc'] = render_doc
    test.globs['get_app'] = get_app


def tearDown(test):
    pass


def create_suite(testfile, layer=None, level=None,
                 tearDown=tearDown, setUp=setUp, cls=doctest.DocFileSuite):
    suite = cls(
        testfile, tearDown=tearDown, setUp=setUp,
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)
    if layer:
        suite.layer = layer
    if level:
        suite.level = level
    return suite


def test_suite():
    s = unittest.TestSuite((
        create_suite('rest.rst'),
        create_suite('validation.rst'),
        create_suite('sphinx/service.txt'),
        create_suite('../../docs/sphinx.txt'),
        create_suite('../../docs/index.txt'),
    ))
    return s



import validictory
import copy


class ValidationException(Exception):
    """A validation exception"""


class Validator(validictory.SchemaValidator):
    """Extend the validator
    """

    def validate_oneOf(self, x, fieldname, schema, props=None):
        """Allows to provide multiple schemas for one property

        If the validation fails the error from the last validation is raised.
        """
        data = x[fieldname]
        error = None
        for s in props:
            try:
                self.validate(data, s)
                return
            except validictory.FieldValidationError as e:
                error = e
        if error:
            raise error


def _number(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


def _boolean(s):
    b = s.lower()
    if b == "true":
        return True
    elif b == "false":
        return False
    raise ValueError

_CONVERSION = {
    "integer": lambda s: int(s),
    "number": _number,
    "boolean": _boolean,
    "array": lambda s: s.split(","),
}


def _convert(kwargs, schema):
    params = copy.deepcopy(kwargs)
    for prop, o in schema.get('properties').iteritems():
        stype = o.get('type')
        if stype in _CONVERSION:
            converter = _CONVERSION[stype]
            v = params.get(prop)
            if v is not None:
                try:
                    params[prop] = converter(v)
                except:
                    # do not convert - validation should fail
                    pass
    return params

def validate(schema, convert_get_params=False):
    """ A decorator to validate the kwargs of the decorated function against
    the given schema. The schema has to be a valid validictory schema.
    If flag 'convert_get_params' is set to True, the kwargs will be
    converted into the appropriate datatypes. Use this flag to validate
    get params correctly on a service.
    """

    def wrapper(f):
        def validation_wrapper(*args, **kwargs):
            if convert_get_params:
                kwargs = _convert(kwargs, schema)
            try:
                validictory.validate(kwargs, schema, Validator)
            except ValueError, error:
                raise ValidationException(error.message)
            return f(*args, **kwargs)
        # link doc with function docstring, so it's accessible for generating
        # documentation
        validation_wrapper.__doc__ = f.__doc__
        return validation_wrapper
    return wrapper


def non_required(schema):
    """ Helper function  to make all properties optional
    """
    s = copy.deepcopy(schema)
    for v in s["properties"].values():
        v['required'] = False
    return s
